package com.mycompany.msgsenderandreceiver.data;


import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlTransient;


@XmlTransient
@XmlAccessorType(XmlAccessType.FIELD)
public class ComparableDTO implements Serializable {
    
    private static final long serialVersionUID = 1L;

    
    @XmlTransient
    private List<String> modifiedFields = new ArrayList<String>();
    
    /**
     * Compares equality of two DTOs. Fields returned by method getExcludedFields
     * are excluded from comparison.
     * 
     * @param other DTO to compare
     * @return true if fields of both DTOs are equal
     */
    public boolean isEqual(Object other) {
        return isEqual(other, true);
    }
    
    public boolean isEqual(Object other, boolean excludeCddFields) {
        
        if(other==null)
        {
            return false;
        }    
        
        modifiedFields.clear();
    
        
        // Make sure that both objects are same class
        if (this.getClass().equals(other.getClass()) == false) {
            return false;
        }
        
        // Get all fields via Reflections
        Field[] fields = getAllFields(getClass());
        
        for (int i = 0; i < fields.length; i++) {
            Field field = fields[i];
              
            try {
                // Make private fields accessible
                field.setAccessible(true);

                if ((field.get(this) == null && field.get(other) != null) || 
                        field.get(this) != null && field.get(other) == null) {
                    // One of the fields is null but the other isn't
                    modifiedFields.add(field.getName());

                } else if (field.get(this) != null && field.get(other) != null) {

                    // Field is DTO collecttion
                    if (field.get(this) instanceof Collection<?>) {
                        Collection thisCollection = (Collection)field.get(this);
                        Collection otherCollection = (Collection)field.get(other);

                        if (thisCollection.size() != otherCollection.size()) {
                            modifiedFields.add(field.getName());
                        } else {

                            for (Object thisItem : thisCollection) {

                                // It is assumed that Collection contains only DTOs that are
                                // inherited from ComparableDTO
                                ComparableDTO thisComparable = (ComparableDTO)thisItem;
                                boolean matchFound = false;

                                for (Object otherItem : otherCollection) {
                                    ComparableDTO otherComparable = (ComparableDTO)otherItem;

                                    if (thisComparable.isEqual(otherComparable)) {
                                        matchFound = true;
                                        break;
                                    }
                                }

                                if (matchFound == false) {
                                    modifiedFields.add(field.getName());
                                    break;
                                }
                            }
                        }

                    // Field is DTO
                    } else if (field.get(this) instanceof ComparableDTO) {

                        
                        ComparableDTO thisComparable = (ComparableDTO)field.get(this);

                        if (thisComparable.isEqual(field.get(other)) == false) {
                            //System.out.println("DEBUG: DTO comparison result is false");
                            modifiedFields.add(field.getName());
                        }

                    // Field is Java basic type or other class than DTO
                    } else {

                        //System.out.println("DEBUG: Field is basic type: " + field.getName());

                        if (field.get(this).equals(field.get(other)) == false) {
                            //System.out.println("DEBUG: Comparison result is false");
                            modifiedFields.add(field.getName());
                        }
                    }
                }   
                    
            } catch (SecurityException ex) {
                System.out.println("isEqual exception" + ex);
            } catch (IllegalAccessException ex) {
                System.out.println("isEqual exception" + ex);
            } catch (IllegalArgumentException ex) {
                System.out.println("isEqual exception" + ex);
            }
            
        }
        
        return getModifiedFieldCount() == 0;
    }

    public List<String> getModifiedFields() {
        return getModifiedFields(true);
    }
    
    public List<String> getModifiedFields(boolean excludeCddFields) {
        List<String> result = new ArrayList<String>();
        
        List<String> excludedFields = getExcludedFields();
        List<String> cddExcludedFields = getCddExcludedFields();

        for (String field : modifiedFields) {
            if (!excludedFields.contains(field)) {
                if (excludeCddFields) {
                    if (!cddExcludedFields.contains(field)) {
                        result.add(field);
                    }
                } else {
                    result.add(field);
                }
            }
        }
        
        return result;
    }
    

    public boolean isFieldModified(String fieldName) {
        if (modifiedFields.contains(fieldName)) {
            return true;
        } else {
            return false;
        }
    }
    
    public int getModifiedFieldCount() {
        return getModifiedFieldCount(true);
    }
    
    public int getModifiedFieldCount(boolean excludeCddFields) {
        return getModifiedFields(excludeCddFields).size();
    }
    
    /**
     * Method should be overriden in DTO and it should return names 
     * of all fields that are not sent to CDD in modify operation.
     * 
     * @return list of names to exclude from comparison
     */
    public List<String> getCddExcludedFields() {
        List<String> fields = new ArrayList<String>();
        fields.add("modifiedFields");
        return fields;
    }
    
    /**
     * Method should be overriden in DTO and it should return names 
     * of all fields that should be excluded from comparison also with
     * 'all' option.
     * 
     * @return list of names to exclude from comparison
     */
    public List<String> getExcludedFields() {
        List<String> fields = new ArrayList<String>();
        fields.add("modifiedFields");
        return fields;
    }
    
    public static Field[] getAllFields(Class clazz) {
        List<Field> results = new ArrayList<>();

        while (true) {
            results.addAll(Arrays.asList(clazz.getDeclaredFields()));

            if (clazz.getSuperclass().getName().startsWith("com.cassidian")) {
                clazz = clazz.getSuperclass();
            } else {
                break;
            }
        }

        return (Field[]) results.toArray(new Field[results.size()]);
    }
}
